unit animcur;

interface

uses Windows, SysUtils, Consts, Classes, Graphics, mmSystem;

type
  JIF = DWORD;
  RTag = packed record //RIFF chunk header
    ckID   : FourCC;
    ckSize : DWORD;
  end;
  TAnimCurHeader = packed record
    cbSizeof  : DWORD;
    cFrames   : DWORD;
    cSteps    : DWORD;
    cx, cy    : DWORD;
    cBitCount : DWORD;
    cPlanes   : DWORD;
    cjifRate  : JIF;
    cflags    : DWORD;
  end;

  PFRAME = ^FRAME;
  FRAME = packed record
    cRef : Integer;          //* reference count */
    iFrame : Integer;
    pfrmNext : PFRAME;
    dwCheckSum : DWORD;     //* checksum of the icon data.  Used to
                           // * compare new imported icons against existing
                           // * ones to see if we can sequence them.
    hcur : HICON;           //* icon representation of raw data below */
    xHotSpot : WORD;
    yHotSpot : WORD;
    Tag : RTag;           //* icon chunk as it appears in file */
    abIcon : ARRAY[0..0] OF Byte;
  end;
  obFrame = class(TObject)
    fFrame : FRAME;
  end;

  TAnimCur = class (TGraphic)
  private
    JifRateList: TList;              // Optional JIFRate info for each step
    FrameOffsets: TList;       // Stream offsets into each frame
    ActualFrames: TList; // frame data
    SeqList: TList;        // Optional frame sequence mapping
    Image: TMemoryStream;      // Memory Image of entire .ANI file
    fAuthor: String;           // Optional author information
    fTitle: String;            // Optional title information
    fHeader: TAnimCurHeader;   // ANI header extracted from file
//    fCurrentJIFs: Integer;     // current JIF count for this step
//    fCurrentStep: Integer;     // current step number
    fCurrentFrame: Integer;    // currently displaying frame number
    fCurrentIcon: hIcon;       // currently displaying icon
    fTransparent: Boolean;     // for transparent blitting 
    fBackColor: TColor;        // background color when not transparent
    procedure Clear;
    procedure SetFrame (Index: Integer);
  protected
    function GetEmpty: Boolean; override;
    function GetHeight: Integer; override;
    function GetWidth: Integer; override;
    procedure SetHeight (Value: Integer); override;
    procedure SetWidth (Value: Integer); override;
  public
    constructor Create; override;
    destructor Destroy; override;
    procedure Assign (Source: TPersistent); override;
    procedure LoadFromStream (Stream: TStream); override;
    procedure SaveToStream (Stream: TStream); override;
//    procedure Animate;
    procedure LoadFromClipboardFormat (AFormat: Word; AData: THandle; APalette: HPalette); override;
    procedure SaveToClipboardFormat (var Format: Word; var Data: THandle; var APalette: HPalette); override;
    procedure Draw (ACanvas: TCanvas; const Rect: TRect); override;
    property Author: String read fAuthor;
    property Title: String read fTitle;
    property Icon: hIcon read fCurrentIcon;
    property Transparent: Boolean read fTransparent write fTransparent default False;
    property BackgroundColor: TColor read fBackColor write fBackColor default clBtnFace;
  end;

implementation



{ TAnimCur }
const
  FourCC_ACON = $4E4F4341; // ACON form type
  FourCC_LIST = $5453494C; // LIST sub-chunk
  FourCC_INFO = $4F464E49; // INFO sub-chunk
  FourCC_INAM = $4D414E49; // INAM - title information
  FourCC_IART = $54524149; // IART - author information
  FourCC_anih = $68696E61; // anih - header information
  FourCC_rate = $65746172; // optional JIF JifRateList sub-chunk
  FourCC_fram = $6D617266; // fram - list of icon frames
  FourCC_icon = $6E6F6369; // icon - start of actual frame
  FourCC_seq  = $20716573; // seq  - optional sequence information
//* If the AF_ICON flag is specified the fields cx, cy, cBitCount, and
// * cPlanes are all unused.  Each frame will be of type ICON and will
// * contain its own dimensional information.
// */

  AF_ICON     = $0001;     //* Windows format icon/cursor animation */
  AF_SEQUENCE = $0002;     //* Animation is sequenced */

function PadUp(D : Dword) : DWord;
begin
  Result := (D+1) AND (NOT 1);
end;

function CalcCheckSum(pb : Pointer; cb : DWord) : DWord;
VAR
  N : Integer;
  I : Int64;
begin
  I := 0;
  FOR N := 0 TO cb-1 DO
    I := I + PByteArray(pb)^[N] AND $FFFFFFFF;
  Result := I;
end;


constructor TAnimCur.Create;
begin
  Inherited Create;
  fTransparent := False;
  fBackColor := clBtnFace;
  JifRateList := TList.Create;
  ActualFrames := TList.Create;
  FrameOffsets := TList.Create;
  SeqList := TList.Create;
  Image := TMemoryStream.Create;
end;

destructor TAnimCur.Destroy;
begin
  Clear;
  Image.Free;
  JifRateList.Free;
  ActualFrames.Free;
  FrameOffsets.Free;
  SeqList.Free;
  Inherited Destroy;
end;

procedure TAnimCur.Clear;
begin
  fAuthor := '--unavailable--';
  fTitle := '--unavailable--';
  Image.Clear;
  JifRateList.Clear;
  FrameOffsets.Clear;
  SeqList.Clear;
  if fCurrentIcon <> 0 then DestroyIcon (fCurrentIcon);
  fCurrentIcon := 0;
end;

procedure TAnimCur.Assign (Source: TPersistent);
begin
  if Source = Nil then Clear
  else if Source is TAnimCur then LoadFromStream (TAnimCur (Source).Image)
  else Inherited Assign (Source);
end;

function TAnimCur.GetEmpty: Boolean;
begin
  Result := FrameOffsets.Count = 0;
end;

procedure TAnimCur.SetHeight (Value: Integer);
begin
  raise EInvalidGraphicOperation.Create (sChangeIconSize);
end;

procedure TAnimCur.SetWidth (Value: Integer);
begin
  raise EInvalidGraphicOperation.Create (sChangeIconSize);
end;

function TAnimCur.GetWidth: Integer;
begin
  Result := fHeader.CX;
end;

function TAnimCur.GetHeight: Integer;
begin
  Result := fHeader.CY;
end;

procedure TAnimCur.LoadFromClipboardFormat (AFormat: Word; AData: THandle; APalette: HPalette);
begin
  raise EInvalidGraphicOperation.Create (sIconToClipboard);
end;

procedure TAnimCur.SaveToClipboardFormat (var Format: Word; var Data: THandle; var APalette: HPalette);
begin
  raise EInvalidGraphicOperation.Create (sIconToClipboard);
end;

procedure TAnimCur.LoadFromStream (Stream: TStream);
var
  ChunkLen: LongInt;
  GotHeader: Boolean;
  Tag : RTag;
  N : Integer;

  procedure FileError;
  begin
    raise Exception.Create ('Animated icon image is not valid');
  end;

  function ReadTag : RTag;
  begin
    Image.ReadBuffer(Result, SizeOf(Result));
  end;

  procedure ReadChunk(Tag : RTag; VAR Data);
  begin
    Image.ReadBuffer(Data, Tag.ckSize);
    IF Odd(Tag.ckSize) THEN
      Image.Seek(1, soFromCurrent);
  end;

  procedure SkipChunk(Tag : RTag);
  begin
    Image.Position := Image.Position + Tag.ckSize;
  end;

  function ReadByte: Byte;
  begin
    Image.ReadBuffer (Result, sizeof (Result));
  end;

  function ReadDWORD : DWORD;
  begin
    Image.ReadBuffer (Result, sizeof (Result));
  end;

  function ReadString: String;
  var
    p: PChar;
    Len: DWORD;
  begin
    Len := ReadDWORD;
    if (Len and 1) <> 0 then Inc (Len);
    GetMem (p, Len + 1);
    p[Len] := #0;
    Image.ReadBuffer (p^, Len);
    Result := StrPas (p);
    FreeMem (p, Len + 1);
  end;

  function ReadStringT(Tag : RTag): String;
  var
    p: PChar;
    Len: DWORD;
  begin
    Len := Tag.ckSize;
    if (Len and 1) <> 0 then Inc (Len);
    GetMem (p, Len + 1);
    p[Len] := #0;
    Image.ReadBuffer (p^, Len);
    Result := StrPas (p);
    FreeMem (p, Len + 1);
  end;

  { Parse the actual icon data itself }
  procedure ParseIconList(Siz : DWord);
  var
      Idx: Integer;
      Len, Next: DWORD;
  begin
    { Discard chunk length }
    if ReadDWORD <> FourCC_fram then FileError;
    { Store frame offsets for later consumption }
    for Idx := 0 to fHeader.cFrames - 1 do begin
        if ReadDWORD <> FourCC_icon then FileError;
        { Save position from beginning of length dword }
        FrameOffsets.Add (Pointer (Image.Position));
        { Read Length of this frame }
        Len := ReadDWORD;
        Next := Len + Image.Position;
        { Dig a little deeper to get the icon size info }
        if Idx = 0 then begin
            Image.Position := Image.Position + 6;
            fHeader.CX := ReadByte;
            fHeader.CY := ReadByte;
        end;

        Image.Position := Next;
    end;
  end;

  procedure ProcessAList(Tag : RTag);
  VAR
    cbChunk : DWORD;
    PF : PFrame;
    N : Integer;
  begin
    cbChunk := PadUp(Tag.ckSize);
    Dec(cbChunk, SizeOf(DWORD));
    CASE ReadDWORD OF
      FourCC_Fram : begin
        WHILE cbChunk >= SizeOf(Tag) DO
          begin
            Tag := ReadTag;
            Dec(cbChunk, SizeOf(Tag));
            IF Tag.ckID = FOURCC_ICON THEN
              begin
                FrameOffsets.Add (Pointer (Image.Position-SizeOf(DWORD)));
                GetMem(PF, SizeOf(Frame) + Tag.ckSize);
                FillChar(PF^, SizeOf(Frame) + Tag.ckSize, 0);
                PF^.cRef := 0;
                Image.ReadBuffer(PF.abIcon, Tag.ckSize);
                pf.dwCheckSum := CalcCheckSum(@pf.abIcon, Tag.ckSize);
                PF.Tag.ckId := FOURCC_Icon;
                PF.Tag.ckSize := Tag.ckSize;
                IF ActualFrames.Count = 0 THEN
                  begin
                    FHeader.CX := PByteArray(@PF.abIcon)^[6];
                    FHeader.CY := PByteArray(@PF.abIcon)^[7];
                  end;
                //TODO: Use check sum, arrange sequence
                ActualFrames.Add(PF);
              end
            ELSE SkipChunk(Tag);
            Dec(cbChunk, PADUP(Tag.ckSize));
          end;
      end;
      FourCC_INFO : begin
        WHILE cbChunk >= SizeOf(Tag) DO
          begin
            Tag := ReadTag;
            Dec(cbChunk, SizeOf(Tag));
            CASE Tag.ckID OF
              FOURCC_INAM : begin
                fTitle := ReadStringT(Tag);
                Dec(cbChunk, PADUP(Tag.ckSize));
              end;
              FOURCC_IART : begin
                fAuthor := ReadStringT(Tag);
                Dec(cbChunk, PADUP(Tag.ckSize));
              end;
              ELSE          begin
                SkipChunk(Tag);
                Dec(cbChunk, PadUP(tag.ckSize));
              end;
            END;
          end;
      end;
      ELSE          begin
        Tag.ckSize := cbChunk;
        SkipChunk(Tag); // skip the remainder of the chunk
      end;
    END;
  end;

begin { LoadFromStream }
  Clear;
  Image.LoadFromStream (Stream);
  GotHeader := False;
  Tag := ReadTag;
  IF Tag.ckID <> FOURCC_RIFF  THEN FileError; // Not valid header
  IF Tag.ckSize > Image.Size  THEN FileError; // Not valid size
  IF ReadDWORD <> FourCC_ACON THEN FileError; // Not ACON chunk

  WHILE Image.Position < Image.Size DO
    begin
      Tag := ReadTag;
      CASE Tag.ckID OF
        FourCC_anih : begin
          IF Tag.ckSize <> SizeOf(fHeader) THEN FileError;
          ReadChunk(Tag, fHeader);
          IF (fHeader.cflags AND AF_ICON = 0) OR (fHeader.cFrames = 0) THEN
            FileError;
          GotHeader := True;
        end;
        FourCC_Rate : begin
          IF Tag.ckSize DIV SizeOf(DWORD) <> fHeader.cSteps THEN
            FileError;
          FOR N := 0 TO fHeader.cSteps-1 DO
            JifRateList.Add(Pointer(ReadDWord));
        end;
        FourCC_seq:begin
          IF Tag.ckSize DIV SizeOf(DWORD) <> fHeader.cSteps THEN
            FileError;
          FOR N := 0 TO fHeader.cSteps-1 DO
            SeqList.Add(Pointer(ReadDWord));
        end;
        FourCC_LIST: ProcessAList(Tag);
        ELSE        SkipChunk(Tag);
      END;
    end;
  SetFrame (0);
end;

procedure TAnimCur.SaveToStream (Stream: TStream);
VAR
  ImageNu : TMemoryStream;
  cbFile, //cbFram,
  cbAuthor, cbTitle, cbInfo : DWord;
//  fRate, fSeq : Boolean;
  N : Integer;

  function Tag(ckID, ckSize : DWORD) : RTag;
  begin
    Result.ckID := ckID;
    Result.ckSize := ckSize;
  end;

  procedure WriteTag(Tag : RTag);
  begin
    ImageNu.WriteBuffer(Tag, SizeOf(Tag));
  end;

  procedure WriteDWORD(D : DWORD);
  begin
    ImageNu.WriteBuffer(D, SizeOf(D));
  end;

  procedure WriteTagData(Tag : RTag; VAR Data);
  VAR cbWrite : DWORD;
  begin
    cbWrite := PadUp(Tag.ckSize);
    WriteTag(Tag);
    ImageNu.WriteBuffer(Data, cbWrite);
  end;

  procedure WriteTagString(Tag : RTag; CONST S : String);
  VAR cbWrite : DWORD;
  begin
    cbWrite := PadUp(Tag.ckSize);
    WriteTag(Tag);
    ImageNu.WriteBuffer(PChar(S)^, cbWrite);
  end;


begin
  if GetEmpty then raise EInvalidGraphicOperation.Create (sInvalidImage);
  ImageNu := TMemoryStream.Create;
  cbFile   := 0;
//  cbFram   := 0;
  cbInfo   := 0;
//  fRate    := False;
//  fSeq     := False;
  cbAuthor := Length(fAuthor);
  cbTitle := Length(fTitle);
  IF (cbAuthor > 0) OR (cbTitle > 0) THEN
    begin
      IF cbTitle > 0 THEN
        begin
          Inc(cbTitle); // for ASCIIZ terminator
          Inc(cbInfo, SizeOf(RTag) + PadUp(cbTitle));
        end;
      IF cbAuthor > 0 THEN
        begin
          Inc(cbAuthor); // for ASCIIZ terminator
          Inc(cbInfo, SizeOf(RTag) + PadUp(cbAuthor));
        end;
      Inc(cbInfo, SizeOf(FourCC));
      Inc(cbFile, SizeOf(RTag) + cbInfo);
    end;
  try
    WriteTag(Tag(FOURCC_RIFF, cbFile));
    WriteDWORD(FourCC_ACON);
    IF (cbAuthor > 0) OR (cbTitle > 0) THEN
      begin
        WriteTag(Tag(FourCC_LIST, cbInfo));
        WriteDWORD(FourCC_INFO);
        IF cbTitle > 0 THEN
          WriteTagString(Tag(FourCC_INAM, cbTitle), fTitle);
        IF cbAuthor > 0 THEN
          WriteTagString(Tag(FourCC_IART, cbAuthor), fAuthor);
      end;
    WriteTagData(Tag(FOURCC_ANIH, SizeOf(fHeader)), fHeader);
    IF JifRateList.Count > 0 THEN
      begin
        WriteTag(Tag(FourCC_rate, fHeader.cSteps*SizeOf(JIF)));
        FOR N := 0 TO fHeader.cSteps-1 DO
          WriteDWORD(DWORD(JifRateList[N]));
      end;
    IF SeqList.Count > 0 THEN
      begin
        WriteTag(Tag(FourCC_Seq, fHeader.cSteps*SizeOf(JIF)));
        FOR N := 0 TO fHeader.cSteps-1 DO
          WriteDWORD(DWORD(SeqList[N]));
      end;

    with ImageNu do Stream.WriteBuffer (Memory^, Size);
//    Image.Clear;
//    Image.WriteBuffer(ImageNu.Memory^, ImageNu.Size);
  finally
    ImageNu.Free;
  end;
end;

procedure TAnimCur.Draw (ACanvas: TCanvas; const Rect: TRect);
var
    bm: TBitmap;
begin
    if fCurrentIcon <> 0 then begin
        if not fTransparent then begin
            bm := TBitmap.Create;
            bm.Width := fHeader.CX;
            bm.Height := fHeader.CY;
            bm.Canvas.Brush.Color := fBackColor;
            bm.Canvas.FillRect (Classes.Rect (0, 0, bm.Width, bm.Height));
            DrawIcon (bm.Canvas.Handle, 0, 0, fCurrentIcon);
            ACanvas.Draw (Rect.Left, Rect.Top, bm);
            bm.Free;
        end else DrawIcon (ACanvas.Handle, Rect.Left, Rect.Top, fCurrentIcon);
    end;
end;

procedure TAnimCur.SetFrame (Index: Integer);
type
    TIconHeader = packed record
        AlwaysZero: Word;
        CursorType: Word;
        NumIcons: Word;
    end;

    TIconDirEntry = packed record
        Width, Height, Colors: Byte;
        Reserved: Byte;
        dwReserved: LongInt;
        dwBytesInRes: LongInt;
        dwImageOffset: LongInt;
    end;

var
    p: PByte;
    ChunkLen: LongInt;
    IconHeader: TIconHeader;
begin
    if (FrameOffsets.Count <> 0) and (Index < fHeader.cFrames) then begin
       fCurrentFrame := Index;
       // Delete any existing icon
       if fCurrentIcon <> 0 then DestroyIcon (fCurrentIcon);
       // Seek to wanted position in stream data
       Image.Position := Integer (FrameOffsets [Index]);
       Image.ReadBuffer (ChunkLen, sizeof (ChunkLen));
       Image.ReadBuffer (IconHeader, sizeof (IconHeader));
       Image.Position := Image.Position + (sizeof (TIconDirEntry) * IconHeader.NumIcons);
       Dec (ChunkLen, sizeof (IconHeader) + (sizeof (TIconDirEntry) * IconHeader.NumIcons));

       p := Image.Memory; Inc (p, Image.Position);
       fCurrentIcon := CreateIconFromResource (p, ChunkLen, True, $30000);
       Changed (Self);
    end;
end;

(*procedure TAnimCur.Animate;
var
    JifRate, NextFrame: Integer;
begin
    if JifRateList.Count = 0 then JifRate := fHeader.cJIFRate else JifRate := Integer (JifRateList [fCurrentStep]);
    Inc (fCurrentJIFs, 4);
    if fCurrentJIFs >= JifRate then begin
        { Time to move on to next step }
        fCurrentJIFs := 0;
        Inc (fCurrentStep);
        if fCurrentStep >= fHeader.cSteps then fCurrentStep := 0;
        if SeqList.Count = 0 then NextFrame := fCurrentFrame + 1 else NextFrame := Integer (SeqList [fCurrentStep]);
        if NextFrame >= fHeader.cFrames then NextFrame := 0;
        if NextFrame <> fCurrentFrame then SetFrame (NextFrame);
    end;
end;*)

end.
