unit IcoObjU;

// Copyright  1999 by Ziff-Davis, Inc.
// Written by Neil J. Rubenking

// This object matches the format of the internal icon-file data
// blocks within an .ANI file. It's also a fair match for the .ICO
// file format, as long as the icon has just a single 32x32 image
// in 16 or 256 colors.

interface
USES Windows, SysUtils, Classes, Graphics, palconsu;

type
  TICONDIR = record
    idReserved : Word;
    idType     : Word;
    idCount    : Word;
  END;

  TIconDirectoryEntry = packed RECORD
    bWidth        : Byte;
    bHeight       : Byte;
    bColorCount   : Byte;
    bReserved     : Byte;
    xHotSpot      : Word;
    yHotSpot      : Word;
    dwBytesInRes  : LongInt;
    dwImageOffset : LongInt;
  End;

  TColBitType  = ARRAY[0..1023] OF Byte;
  TMaskBitType = ARRAY[0..127] OF Byte;

  TInternalCur = class(TObject)
  private
      fValid    : Boolean;
      fExcess   : Boolean;
      fID       : TIconDir;
      fIDE      : TIconDirectoryEntry;
      fBIH      : TBitmapInfoheader;
      fPal      : TPalArray;
      fcolbits  : TColBitType;
      fmaskbits : TMaskBitType;
      fIcon     : TIcon;
      fDataSize : Integer;
      fTag      : Integer;
      fBlackIdx : Integer; // used for transparent
      fWhiteIdx : Integer; // used for inverted
    function GetAsIcon : TIcon;
    function GetIs16Color: Boolean;
    function GetColBits(aCol,aRow: Integer): Integer;
    procedure SetColBits(aCol,aRow : Integer; const Value: Integer);
    function GetMaskBits(aCol,aRow: Integer): Integer;
    procedure SetMaskBits(aCol,aRow : Integer; const Value: Integer);
    function GetBlackIdx: Integer;
    function GetWhiteIdx: Integer;
    public
      constructor Create; overload;
      constructor Create(Stream : TStream); overload;
      constructor Create(orig : TInternalCur); overload;
      constructor Create(orig: TInternalCur; vtag : Integer);
        overload;
      constructor Create(is16color: Boolean; xHot, yHot : Word;
        const vpal : TPalArray); overload;
      destructor Destroy; override;
      procedure LoadFromStream(Stream : TStream);
      procedure SaveToStream(Stream : TStream);
      procedure LoadFromFile(const fileName : String);
      procedure SaveToFile(const fileName : String);
      function Same(Oth : TInternalCur) : Boolean;
      // read-only properties
      property DataSize  : Integer read fDataSize;
      property AsIcon    : TIcon read GetAsIcon;
      property Valid     : Boolean Read fValid;
      property Palette   : TPalArray read fPal;
      property Is16color : Boolean read GetIs16Color;
      property Excess    : Boolean read fExcess;
      property BlackIdx  : Integer read GetBlackIdx;
      property WhiteIdx  : Integer read GetWhiteIdx;
      // read-write properties
      property Tag  : Integer read fTag write fTag;
      property HotX : Word read fIDE.xHotspot write fIDE.xHotspot;
      property HotY : Word read fIDE.yHotspot write fIDE.yHotspot;
      property ColBits[aCol, aRow : Integer] : Integer
        read GetColBits write SetColBits;
      property MaskBits[aCol, aRow : Integer] : Integer
        read GetMaskBits write SetMaskBits;
  end;

implementation
{ TInternalCur }


constructor TInternalCur.Create;
// Just initialize; called by other constructors
begin
  Inherited Create;
  fValid     := False;
  fExcess    := False;
  fIcon      := nil;
  fBlackIdx  := -1;
  fWhiteIdx  := -1;
  Tag        := 0;
  FillChar(fID,       SizeOf(fID), 0);
  FillChar(fIDE,      SizeOf(fIDE), 0);
  FillChar(fBIH,      SizeOf(fBIH), 0);
  FillChar(fPal,      SizeOf(fPal), 0);
  FillChar(fcolbits,  SizeOf(fcolbits), 0);
  FillChar(fmaskbits, SizeOf(fmaskbits), 0);
end;

constructor TInternalCur.Create(Stream: TStream);
// Create and load from stream in one operation
begin
  Create;
  LoadFromStream(Stream);
  fBlackIdx  := -1;
  fWhiteIdx  := -1;

end;

constructor TInternalCur.Create(orig: TInternalCur);
// Create clone of existing object
begin
  Create;
  fValid    :=  orig.fValid;
  fExcess   :=  orig.fExcess;
  fID       :=  orig.fID;
  fIDE      :=  orig.fIDE;
  fBIH      :=  orig.fBIH;
  fPal      :=  orig.fPal;
  fcolbits  :=  orig.fcolbits;
  fmaskbits :=  orig.fmaskbits;
  fIcon     :=  nil;;
  fDataSize :=  orig.fDataSize;
  fTag      :=  orig.fTag;
end;

constructor TInternalCur.Create(orig: TInternalCur; vtag: Integer);
// Clone and set tag
begin
  Create(orig);
  Tag := vTag;
end;

constructor TInternalCur.Create(is16color: Boolean; xHot,
  yHot : Word; const vpal: TPalArray);
// Create a fully functional blank frame
begin
  Create;
  fID.idType    := 2;
  fID.idCount   := 1;
  fIDE.bWidth   := 32;
  fIDE.bHeight  := 32;
  fIDE.xHotSpot := xHot;
  fIDE.yHotSpot := yHot;
  fDataSize := SizeOf(fID) + SizeOf(fIDE) + SizeOf(fBIH) +
    SizeOf(fMaskBits);
  IF is16Color THEN
    begin
      fIDE.dwBytesInRes := 744;
      fBIH.biClrUsed    := 0;
      fBIH.biBitCount   := 4;
      Inc(fDataSize, 16*SizeOf(TColorRef));
      Inc(fDataSize, 512);
    end
  ELSE
    begin
      fIDE.dwBytesInRes := 2216;
      fBIH.biClrUsed    := 256;
      fBIH.biBitCount   := 8;
      Inc(fDataSize, 256*SizeOf(TColorRef));
      Inc(fDataSize, 1024);
    end;
  fIDE.dwImageOffset := 22;
  fBIH.biSize        := SizeOf(fBIH);
  fBIH.biWidth       := 32;
  fBIH.biHeight      := 64;
  fBIH.biPlanes      := 1;
  fBIH.biSizeImage   := 128+(1024*fBIH.biBitCount DIV 8);
  MOVE(vPal, fPal, SizeOf(fPal));
  fValid := True;
end;

destructor TInternalCur.Destroy;
begin
  IF Assigned(fIcon) THEN fIcon.Free;
  Inherited Destroy;
end;

procedure TInternalCur.LoadFromStream(Stream: TStream);
VAR
  N  : Integer;
  RQ : TRGBQuad;
  P  : Integer;
begin
  P := Stream.Position;
  try
    Stream.ReadBuffer(fID, SizeOf(fID));
    IF fID.idCount > 1 THEN
      begin
        fID.idCount := 1;
        fExcess     := True;
      end;
    Stream.ReadBuffer(fIDE, SizeOf(fIDE));
    IF fIDE.dwImageOffset <> 22 THEN
      begin
        Stream.Seek(P+fIDE.dwImageOffset, soFromBeginning);
        fIDE.dwImageOffset := 22;
      end;
    Stream.ReadBuffer(fBIH, SizeOf(fBIH));
    IF (fBIH.biBitCount <> 4) AND (fBIH.biBitCount <> 8) THEN
      Raise(Exception.Create('Can only load 16-color or '+
        '256-color cursors.'));
    FOR N := 0 TO (1 SHL fBIH.biBitCount)-1 DO
      begin
        Stream.ReadBuffer(RQ, SizeOf(RQ));
        WITH RQ DO fPal[N] := RGB(RGBRed, RGBGreen, RGBBlue);
      end;
    Stream.ReadBuffer(fcolbits, 32*32*fBIH.biBitCount DIV 8);
    Stream.ReadBuffer(fmaskbits, 32*32 DIV 8);
    fDataSize := Stream.Position - P;
    IF fID.idReserved   <> 0 THEN Exit;
    IF fIDE.bWidth      <> 32 THEN Exit;
    IF fIDE.bHeight     <> 32 THEN Exit;
    fValid := True;
  except
    ON EReadError DO;
  end;
end;

procedure TInternalCur.SaveToStream(Stream: TStream);
VAR
  N  : Integer;
  RQ : TRGBQuad;
begin
  IF NOT fValid THEN
    Raise(Exception.Create('No icon data loaded'));
  try
    Stream.WriteBuffer(fID, SizeOf(fID));
    Stream.WriteBuffer(fIDE, SizeOf(fIDE));
    Stream.WriteBuffer(fBIH, SizeOf(fBIH));
    FOR N := 0 TO (1 SHL fBIH.biBitCount)-1 DO
      begin
        RQ.rgbRed       := GetRValue(fPal[N]);
        RQ.rgbGreen     := GetGValue(fPal[N]);
        RQ.rgbBlue      := GetBValue(fPal[N]);
        RQ.rgbReserved  := 0;
        Stream.WriteBuffer(RQ, SizeOf(RQ));
      end;
    Stream.WriteBuffer(fcolbits, 32*32*fBIH.biBitCount DIV 8);
    Stream.WriteBuffer(fmaskbits, 32*32 DIV 8);
  except
    ON EWriteError DO;
  end;
end;

procedure TInternalCur.LoadFromFile(const fileName: String);
VAR Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead OR
    fmShareDenyWrite);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TInternalCur.SaveToFile(const fileName: String);
VAR Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmCreate);
  try
    SaveToStream(Stream);
  finally
    Stream.Free;
  end;
end;

function TInternalCur.Same(Oth: TInternalCur): Boolean;
VAR TM1, TM2 : TMemoryStream;
begin
  TM1 := TMemoryStream.Create;
  TM2 := TMemoryStream.Create;
  try
    SaveToStream(TM1);
    Oth.SaveToStream(TM2);
    Result := CompareMem(TM1.Memory, TM2.Memory, TM1.Size);
  finally
    TM1.Free;
    TM2.Free;
  end;
end;

function TInternalCur.GetAsIcon : TIcon;
VAR
  TMS : TMemoryStream;
  Was : Integer;
begin
  IF NOT Assigned(fIcon) THEN
    begin
    fIcon := TIcon.Create;
    TMS   := TMemoryStream.Create;
    try
      Was := fID.idType;
      fID.idType := 1; // set it to icon type
      SaveToStream(TMS);
      fID.idType := Was;
      TMS.Seek(0, soFromBeginning);
      fIcon.LoadFromStream(TMS);
    finally
      TMS.Free;
    end;
  end;
  Result := fIcon;
end;

function TInternalCur.GetIs16Color: Boolean;
begin
  IF NOT fValid THEN
    Raise(Exception.Create('No icon data loaded'));
  Result := fBIH.biBitCount = 4;
end;

function TInternalCur.GetColBits(aCol, aRow: Integer): Integer;
begin
  IF NOT fValid THEN
    Raise(Exception.Create('No icon data loaded'));
  IF (aCol < 0) OR (aRow < 0) OR (aCol > 31) OR (aRow > 31) THEN
    Raise(Exception.Create('Pixel index out of range'));
  IF fBIH.biBitCount = 4 THEN
    begin
      IF Odd(aCol) THEN
        Result := fColBits[aRow*16+(aCol DIV 2)] AND $F
      ELSE
        Result := fColBits[aRow*16+(aCol DIV 2)] SHR 4;
    end
  ELSE IF fBIH.biBitCount = 8 THEN
    Result := fColBits[aRow*32+aCol]
  ELSE Result := 0;
end;

procedure TInternalCur.SetColBits(aCol, aRow : Integer;
  const Value: Integer);
begin
  IF (aCol < 0) OR (aRow < 0) OR (aCol > 31) OR (aRow > 31) THEN
    Raise(Exception.Create('Pixel index out of range'));
  IF fBIH.biBitCount = 4 THEN
    begin
      IF Odd(aCol) THEN
        fColBits[aRow*16+(aCol DIV 2)] :=
          (fColBits[aRow*16+(aCol DIV 2)] AND $F0) OR
            (Value AND $F)
      ELSE
        fColBits[aRow*16+(aCol DIV 2)] :=
          (fColBits[aRow*16+(aCol DIV 2)] AND $F) OR
            ((Value AND $F) SHL 4);
    end
  ELSE IF fBIH.biBitCount = 8 THEN
    fColBits[aRow*32+aCol] := Value;
end;

function TInternalCur.GetMaskBits(aCol, aRow: Integer): Integer;
VAR idx, bit : Integer;
begin
  IF NOT fValid THEN
    Raise(Exception.Create('No icon data loaded'));
  IF (aCol < 0) OR (aRow < 0) OR (aCol > 31) OR (aRow > 31) THEN
    Raise(Exception.Create('Pixel index out of range'));
  idx := (aRow*4) + (aCol DIV 8);
  bit := 1 SHL (7-(aCol MOD 8));
  IF fmaskBits[idx] AND bit > 0 THEN Result := 1
  ELSE                               Result := 0;
end;

procedure TInternalCur.SetMaskBits(aCol, aRow : Integer;
  const Value: Integer);
VAR
  idx, bit : Integer;
begin
  IF (aCol < 0) OR (aRow < 0) OR (aCol > 31) OR (aRow > 31) THEN
    Raise(Exception.Create('Pixel index out of range'));
  idx := (aRow*4) + (aCol DIV 8);
  bit := 1 SHL (7-(aCol MOD 8));
  IF value = 1 THEN fMaskBits[idx] := fMaskBits[idx] OR bit
  ELSE              fMaskBits[idx] := fMaskBits[idx] AND (NOT bit);
end;

function TInternalCur.GetBlackIdx: Integer;
VAR N : Integer;
begin
  IF fBlackIdx = -1 THEN
    begin
      FOR N := 0 TO (1 SHL fBIH.biBitCount) - 1 DO
        IF fPal[N] AND $FFFFFF = $000000 THEN
          begin
            fBlackIdx := N;
            Break;
          end;
    end;
  IF fBlackIdx = -1 THEN fBlackIdx := 0;
  Result := fBlackIdx;
end;

function TInternalCur.GetWhiteIdx: Integer;
VAR N : Integer;
begin
  IF fWhiteIdx = -1 THEN
    begin
      FOR N := 0 TO (1 SHL fBIH.biBitCount) - 1 DO
        IF fPal[N] AND $FFFFFF = $FFFFFF THEN
          begin
            fWhiteIdx := N;
            Break;
          end;
    end;
  IF fWhiteIdx = -1 THEN fWhiteIdx := (1 SHL fBIH.biBitCount) - 1;
  Result := fWhiteIdx;
end;

end.
