unit palfrmu;

// Copyright  1999 by Ziff-Davis, Inc.
// Written by Neil J. Rubenking

// This Delphi 5 TFrame descendant encapsulates the palette
// color-picker, thereby keeping its complexity out of the
// main form

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, palconsu, StdCtrls;

type
  TPalFrame = class(TFrame)
    pnlPal : TPanel;
    pbPal  : TPaintBox;
    procedure pbPalMouseMove(Sender: TObject; Shift: TShiftState;
      X, Y: Integer);
    procedure pbPalMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure pbPalPaint(Sender: TObject);
  private
    { Private declarations }
    PalM              : Integer; // margin for palette display
    thePalColor       : Integer; // under-cursor palette color
    fForeColor        : Integer; // foreground color
    fBackColor        : Integer; // background color
    palR              : Integer; // # rows in palette display
    palC              : Integer; // # cols in palette display
    palS              : Integer; // cell size in palette display
    palW              : Integer; // pixel width for palette display
    palH              : Integer; // pixel height for palette display
    IsPal16           : Boolean;
    aPalette          : TPalArray; // array of palette entries
    PaletteObj        : HPalette;  // palette handle from aPalette

    procedure SetAPalette(value : TPalArray);
  public
    { Public declarations }
    OnColorChange : TNotifyEvent;
    procedure SetPalValues(is16Pal: Boolean);
    property Pal : TPalArray read aPalette write SetAPalette;
    property PalO : HPalette read PaletteObj;
    property fore : Integer read fForeColor write fForeColor;
    property back : Integer read fBackColor write fBackColor;
  end;

implementation

{$R *.DFM}
const
  palMaxim = 255;
  palTrans = 256;
  palInvrt = 257;

procedure TPalFrame.pbPalMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
VAR Xcell, Ycell : Integer;
begin
  thePalColor               := -1; // default - no color
  pbPal.Hint := ''; // default - no status
  Xcell                        := (X-palM) DIV palS;
  Ycell                        := (Y-palM) DIV palS;
  IF X < PalM      THEN Exit; // outside, to left
  IF Y < PalM      THEN Exit; // outside, to top
  IF Ycell >= palR THEN Exit; // outside, to bottom
  IF Xcell > palC  THEN Exit; // outside, to right
  IF Xcell = palC  THEN       // in last "special" column
    begin
      IF Ycell = 0 THEN
        begin
          pbPal.Hint  := 'transparent';
          thePalColor := palTrans;
        end
      ELSE IF Ycell = palR-1 THEN
        begin
          pbPal.Hint  := 'inverted';
          thePalColor := palInvrt;
         end;
      Exit;
    end;
  //(XCell, YCell) is within the palette rectangle
  thePalColor := Ycell*palC+Xcell;
  pbPal.Hint := StringForColor(aPalette, thePalColor);
end;

procedure TPalFrame.pbPalMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  IF thePalColor < 0 THEN Exit;
  CASE Button OF
    mbLeft  : IF thePalColor <> fForeColor THEN
      begin
        fForeColor := thePalColor;
        IF Assigned(OnColorChange) THEN OnColorChange(Self);
      end;
    mbRight : IF thePalColor <> fBackColor THEN
      begin
        fBackColor := thePalColor;
        IF Assigned(OnColorChange) THEN OnColorChange(Self);
      end;
  end;
end;

procedure TPalFrame.pbPalPaint(Sender: TObject);
VAR co, ro, N, M1 : Integer;
begin
  M1 := PalM-1;
  WITH pbPal.Canvas DO
    begin
      SelectPalette(Handle, PaletteObj, FALSE);
      RealizePalette(Handle);
      Pen.Color := clBtnShadow;
      // Draw the vertical lines
      FOR N := 0 TO palC+1 DO
        begin
          MoveTo((M1)+(N*palS), (M1));
          LineTo((M1)+(N*palS), (M1)+palS);
        end;
      // Draw the horizontal lines. Top and bottom lines as well
      // as those just "in" from top and bottom are one cell longer
      FOR N := 0 TO palR DO
        begin
          MoveTo((M1), (M1)+(N*palS));
          IF (N > 1) AND (N < palR-1) THEN
            LineTo((M1)+palW, (M1)+(N*palS))
          ELSE
            LineTo((M1)+palW+palS, (M1)+(N*palS))
        end;
      // Draw the ends of the extra cells (transparent and inverted)
      MoveTo(M1+(palC+1)*palS, (M1));
      LineTo(M1+(palC+1)*palS, (M1)+palS);
      MoveTo(M1+(palC+1)*palS, (M1)+palH-palS);
      LineTo(M1+(palC+1)*palS, (M1)+palH);
      // Fill in the palette colors
      FOR ro := 0 TO palR-1 DO
        FOR co := 0 TO palC-1 DO
          begin
            Brush.Color := aPalette[(ro*palC)+co];
            FillRect(Bounds(PalM+(co*palS),
              PalM+(ro*palS), palS-1, palS-1));
            end;
      // Draw blue backslash for transparent
      Pen.Color := clBlue;
      MoveTo(PalM+palC*palS+1, PalM+1);
      LineTo(PalM+(palC+1)*palS-2, PalM+palS-2);
      // Draw red slash for inverted
      Pen.Color := clRed;
      MoveTo(PalM+palC*palS+1, PalM+palH-4);
      LineTo(PalM+(palC+1)*palS-2, PalM+palH-palS-1);
    end;
end;

procedure TPalFrame.SetAPalette(value: TPalArray);
VAR N : Integer;

  function GetPalFrom(P : TPalArray; colors : Integer) : HPalette;
  VAR
    PLP : PLogPalette;
    N   : Integer;
  begin
    GetMem(PLP, SizeOf(PLP)+Colors*SizeOf(TPaletteEntry));
    try
      PLP^.palVersion := $300;
      PLP^.palNumEntries := colors;
      {$IFOPT R+}
      {$R-}
      {$DEFINE WasPlus}
      {$ENDIF}
      FOR N := 0 TO colors-1 DO
        WITH PLP^.palPalEntry[N] DO
          begin
            peRed    := GetRValue(P[N]);
            peGreen  := GetGValue(P[N]);
            peBlue   := GetBValue(P[N]);
            peFlags  := 0;//PC_NOCOLLAPSE;
          end;
      {$IFDEF WasPlus}
      {$R+}
      {$ENDIF}
      Result := CreatePalette(PLP^);
    finally
      FreeMem(PLP);
    end;
  end;
begin
  aPalette := Value;
  FOR N := 0 TO 255 DO // Helps if forced into 256-color
    aPalette[N] := aPalette[N]+$02000000;
  fForeColor := 0;
  IF isPal16 THEN fBackColor := 15
  ELSE fBackColor := 255;
  PaletteObj := GetPalFrom(aPalette, fBackColor+1);
  // If first palette color is not black, seek black
  IF aPalette[fForeColor] <> $000000 THEN
    FOR N := 0 TO fBackColor DO
      IF aPalette[N] = $000000 THEN
        begin
          fForeColor := N;
          Break;
        end;
  // If last palette color is not white, seek white
  IF aPalette[fBackColor] <> $FFFFFF THEN
    FOR N := fBackColor DOWNTO 0 DO
      IF aPalette[N] = $FFFFFF THEN
        begin
          fBackColor := N;
          Break;
        end;
  pbPal.Refresh;
  IF Assigned(OnColorChange) THEN OnColorChange(Self);
end;

procedure TPalFrame.SetPalValues(is16Pal: Boolean);
begin
  PalM       := 3;
  IsPal16    := is16Pal;
  fForeColor := 0;
  IF is16Pal THEN
    begin
      palR       := 2;
      palC       := 8;
      palS       := 16;
      SetAPalette(DefPal16);
      fbackColor := 15;
    end
  ELSE
    begin
      palR       := 4;
      palC       := 64;
      palS       := 9;
      SetAPalette(DefPal256);
      fbackColor := 255;
    end;
  palW := (palS*palC)+1;
  palH := (palS*palR)+1;
  pbPal.Refresh;
  IF Assigned(OnColorChange) THEN OnColorChange(Self);
end;

end.
