unit readAniU;

// Copyright  1999 by Ziff-Davis, Inc.
// Written by Neil J. Rubenking

interface

uses Windows, mmSystem, SysUtils, Classes, Graphics, icoObjU,
  PalConsU;

type
  JIF = DWORD;
  RTag = packed record //RIFF chunk header
    ckID   : FourCC;
    ckSize : DWORD;
  end;

type
  TAniCurHeader = packed record
    cbSizeof  : DWORD;
    cFrames   : DWORD;
    cSteps    : DWORD;
    cx, cy    : DWORD;
    cBitCount : DWORD;
    cPlanes   : DWORD;
    cjifRate  : JIF;
    cflags    : DWORD;
  end;

  TAniCursor = class(Tobject)
    private
      fHeader    : TAniCurHeader;
      fAuthor    : String;
      fTitle     : String;
      fJifRates  : TList;
      fSequences : TList;
      fFrams    : TList;
      fExcess    : Boolean;  // Excess data within .ani file
      procedure FErrIn;
      procedure FErrOut;
      function GetFrameCount : Integer;
      function GetHotX       : WORD;
      function GetHotY       : WORD;
      function GetIsPal16    : Boolean;
      function GetRateCount  : Integer;
      function GetSeqCount   : Integer;
      function GetPalette    : TPalArray;
      function GetFrame(Index: Integer)    : TInternalCur;
      function GetRate(Index: Integer)     : DWORD;
      function GetSequence(Index: Integer) : DWORD;
    public
      constructor Create; overload;
      constructor Create(const filename : String); overload;
      constructor Create(const vTitle, vAuthor : String;
        vAllJifs : DWORD; is16color : Boolean); overload;
      destructor Destroy; override;
      procedure LoadFromStream(Stream : TStream);
      procedure SaveToStream(Stream : TStream);
      procedure LoadFromFile(const fileName : String);
      procedure SaveToFile(const fileName : String);
      procedure AddImage(T : TInternalCur);
      procedure PrepareToSave(vHotX, vHotY : Word);
      //read-only properties
      property Excess     : Boolean       read fExcess;
      property Header     : TAniCurHeader read fHeader;
      property HotX       : WORD          read GetHotX;
      property HotY       : WORD          read GetHotY;
      property IsPal16    : Boolean       read GetIsPal16;
      property FrameCount : Integer       read GetFrameCount;
      property RateCount  : Integer       read GetRateCount;
      property SeqCount   : Integer       read GetSeqCount;
      property Palette    : TPalArray     read GetPalette;
      // read-only array properties
      property Frames[Index: Integer]: TInternalCur read GetFrame;
      property Rates[Index: Integer]: DWORD read GetRate;
      property Sequences[Index: Integer]: DWORD read GetSequence;
      //read-write properties
      property Author : String read fAuthor write fAuthor;
      property Title  : String read fTitle write fTitle;
  end;

implementation

const
  // RIFF chunk type headers
  FOURCC_ACON = $4E4F4341; // 'A','C','O','N'
  FOURCC_anih = $68696E61; // 'a','n','i','h'
  FOURCC_rate = $65746172; // 'r','a','t','e'
  FOURCC_seq  = $20716573; // 's','e','q',' '
  FOURCC_fram = $6D617266; // 'f','r','a','m'
  FOURCC_icon = $6E6F6369; // 'i','c','o','n'
  FOURCC_INFO = $4F464E49; // 'I','N','F','O'
  FOURCC_IART = $54524149; // 'I','A','R','T'
  FOURCC_INAM = $4D414E49; // 'I','N','A','M'

//"If the AF_ICON flag is specified the fields cx, cy, cBitCount, and
// cPlanes are all unused.  Each frame will be of type ICON and will
// contain its own dimensional information.

  AF_ICON     = $0001; // Windows format icon/cursor animation
  AF_SEQUENCE = $0002; // Animation is sequenced

function PadUp(D : Dword) : DWord;
begin
  Result := (D+1) AND (NOT 1);
end;

function ReadTag(Stream : TStream; VAR tag : RTag) : Boolean;
begin
  try
    Stream.ReadBuffer(Tag, SizeOf(Tag));
    Result := True;
  except
    On Exception DO Result := False;
  end;
end;

function ReadChunk(Stream : TStream; VAR Tag : RTag;
  VAR pv) : Boolean;
begin
  try
    Stream.ReadBuffer(pv, Tag.ckSize);
    // Align to word boundary
    IF Odd(tag.ckSize) THEN Stream.Seek(1, soFromCurrent);
    Result := True;
  except
    On Exception DO Result := False;
  end;
end;

function ReadChunkS(Stream : TStream; VAR Tag : RTag;
  VAR S : String) : Boolean;
VAR buffer : PChar;
begin
  try
    GetMem(Buffer, Tag.ckSize+1);
    FillChar(Buffer^, Tag.ckSize+1, 0);
    Stream.ReadBuffer(Buffer^, Tag.ckSize);
    S := StrPas(Buffer);
    // Align to word boundary
    IF Odd(tag.ckSize) THEN Stream.Seek(1, soFromCurrent);
    Result := True;
  except
    On Exception DO Result := False;
  end;
end;

function SkipChunk(Stream : TStream; VAR Tag : RTag) : Boolean;
begin
  try
    Result := True;
    Stream.Seek(PadUp(Tag.ckSize), soFromCurrent);
  except
    ON Exception DO Result := False;
  end;
end;

function WriteTag(Stream : TStream; tag : RTag) : Boolean;
begin
  try
    Stream.WriteBuffer(Tag, SizeOf(Tag));
    Result := True;
  except
    On Exception DO Result := False;
  end;
end;

function WriteTagData(Stream : TStream; tag : RTag;
  VAR Data) : Boolean;
const zero : Byte = 0;
begin
  try
    Stream.WriteBuffer(Tag, SizeOf(Tag));
    Stream.WriteBuffer(Data, Tag.ckSize);
    // Align to word boundary
    IF Odd(Tag.ckSize) THEN Stream.WriteBuffer(Zero, 1);
    Result := True;
  except
    On Exception DO Result := False;
  end;
end;

function WriteTagString(Stream : TStream; tag : RTag;
  const S : String) : Boolean;
VAR buffer : PChar;
begin
  GetMem(Buffer, Length(S)+1);
  try
    FillChar(Buffer^, Length(S)+1, 0);
    StrPCopy(Buffer, S);
    Result := WriteTagData(Stream, Tag, buffer^);
  finally
    FreeMem(Buffer);
  end;
end;

function WriteType(Stream : TStream; ckID : DWORD) : Boolean;
begin
  try
    Stream.WriteBuffer(ckID, SizeOf(ckID));
    Result := True;
  except
    On Exception DO Result := False;
  end;
end;

{ TAniCursor }

constructor TAniCursor.Create;
// Create a tabula rasa
begin
  Inherited Create;
  fExcess := False;
  fAuthor := '';
  fTitle  := '';
  FillChar(fHeader, SizeOf(fHeader), 0);
  WITH fHeader DO
    begin
      cbSizeof  := SizeOf(fHeader);
      cPlanes   := 1;
      cjifRate  := 10;
      cflags    := 1;
    end;
  fJifRates  := TList.Create;
  fSequences := TList.Create;
  fFrams    := TList.Create;
end;

constructor TAniCursor.Create(const filename: String);
// Create and load from a file
begin
  Create;
  LoadFromFile(filename);
end;

constructor TAniCursor.Create(const vTitle, vAuthor: String;
  vAllJifs: DWORD; is16color : Boolean);
// Create, ready to have images added with AddImage
begin
  Create;
  fTitle  := vTitle;
  fAuthor := vAuthor;
  fHeader.cjifRate := vAllJifs;
  IF Is16color THEN
    fHeader.cBitCount := 0
  ELSE fHeader.cBitCount := 4;
end;

destructor TAniCursor.Destroy;
begin
  fJifRates.Free;
  fSequences.Free;
  fFrams.Free;
  Inherited Destroy;
end;

procedure TAniCursor.LoadFromStream(Stream: TStream);
// Animated cursor file has this format:
//   RIFF( 'ACON'
//       LIST( 'INFO'
//           INAM( <name> )
//           IART( <artist> )
//       )
//       'anih' ( <anihdr> )
//       ['rate'( <rateinfo> )  ]
//       ['seq '( <seq_info> )]
//       LIST( 'fram' icon( <icon_file> ) )
//   )
VAR
  Tag     : RTag;
  N       : Integer;
  P       : DWORD;
  cbChunk : DWORD;
  TII     : TInternalCur;

  function ReadDWORD : DWord;
  begin
    try
      Stream.ReadBuffer(Result, SizeOf(Result));
    except
      ON Exception DO FErrIn;
    end;
  end;

begin
  // Clear out any existing data
  fJifRates.Clear;
  fSequences.Clear;
  IF fFrams.Count > 0 THEN
    FOR N := 0 TO fFrams.Count-1 DO
      WITH TInternalCur(fFrams[N]) DO
        Free;
  fFrams.Clear;
  fAuthor := '';
  fTitle := '';
  IF NOT ReadTag(Stream, Tag) THEN FErrIn;
  IF Tag.ckID <> FOURCC_RIFF  THEN FErrIn; // Not valid header
  IF Integer(Tag.ckSize) > Stream.Size  THEN
    FErrIn; // Not valid size
  IF ReadDWORD <> FourCC_ACON THEN FErrIn; // Not ACON chunk
  WHILE Stream.Position < Stream.Size DO
    begin
      IF NOT ReadTag(Stream, Tag) THEN FErrIn;
      CASE Tag.ckID OF
        FourCC_anih : begin
          IF Tag.ckSize <> SizeOf(fHeader) THEN FErrIn;
          IF NOT ReadChunk(Stream, Tag, fHeader) THEN FErrIn;
          IF (fHeader.cflags AND AF_ICON = 0) OR
             (fHeader.cFrames = 0) THEN
            FErrIn;
        end;
        FourCC_Rate : begin
          IF Tag.ckSize DIV SizeOf(DWORD) <> fHeader.cSteps THEN
            FErrIn;
          FOR N := 0 TO fHeader.cSteps-1 DO
            fJifRates.Add(Pointer(ReadDWord));
        end;
        FourCC_seq  : begin
          IF Tag.ckSize DIV SizeOf(DWORD) <> fHeader.cSteps THEN
            FErrIn;
          FOR N := 0 TO fHeader.cSteps-1 DO
            fSequences.Add(Pointer(ReadDWord));
        end;
        FourCC_LIST : begin
          cbChunk  := PadUp(Tag.ckSize);
          Tag.ckID := ReadDWORD;
          Dec(cbChunk, SizeOf(Tag.ckID));
          CASE Tag.ckID OF
            FOURCC_fram : begin // The actual icon-data
              WHILE cbChunk >= SizeOf(Tag) DO
                begin
                  IF NOT ReadTag(Stream, Tag) THEN FErrIn;
                  Dec(cbChunk, SizeOf(Tag));
                  CASE Tag.ckID OF
                    FOURCC_Icon : begin
                      P := Stream.Position;
                      TII := TInternalCur.Create(Stream);
                      IF TII.Excess THEN fExcess := True;
                      fFrams.Add(TII);
                      Stream.Seek(P+PadUp(Tag.ckSize),
                        soFromBeginning);
                    end;
                    ELSE SkipChunk(Stream, Tag);
                  END;
                  Dec(cbChunk, PadUp(Tag.ckSize));
                end;
            end;
            FOURCC_INFO : begin // INFO includes author and title
              WHILE cbChunk >= SizeOf(Tag) DO
                begin
                  IF NOT ReadTag(Stream, Tag) THEN FErrIn;
                  Dec(cbChunk, SizeOf(Tag));
                  CASE Tag.ckID OF
                    FOURCC_INAM : IF NOT ReadChunkS(Stream, Tag,
                      fTitle) THEN FErrIn;
                    FOURCC_IART : IF NOT ReadChunkS(Stream, Tag,
                      fAuthor) THEN FErrIn;
                    ELSE          SkipChunk(Stream, Tag);
                  END;
                  Dec(cbChunk, PadUp(Tag.ckSize));
                end;
            end;
            ELSE SkipChunk(Stream, Tag);
          END;
        end;
        ELSE SkipChunk(Stream, Tag);
      END;
    end;
end;

procedure TAniCursor.SaveToStream(Stream: TStream);
VAR
  i        : Integer;
  cbFile   : DWORD;
  cbFram   : DWORD;
  cbINFO   : DWORD;
  cbTitle  : DWORD;
  cbAuthor : DWORD;
  fRate    : Boolean; // true if rate-list used
  fSeq     : Boolean; // true if sequence-list used
  Tag      : RTag;
begin
  fRate    := fJifRates.Count > 0;
  fSeq     := fSequences.Count > 0;
  cbFram   := 0;
  cbINFO   := 0;
  cbTitle  := Length(fTitle);
  cbAuthor := Length(fAuthor);
  FOR i := 0 TO fFrams.Count-1 DO
    begin
      Inc(cbFram, SizeOf(RTag));
      Inc(cbFram, TInternalCur(fFrams[i]).DataSize);
    end;
  // At this point, cbFram = the size required by all the frames,
  // add in the rate, seq, anih, and INFO list sizes as well as
  // all the other overhead.

  Inc(cbFram, SizeOf(FOURCC));  // fram type
  cbFile := cbFram;
  cbFile := cbFile + sizeof(FOURCC) +    //ACON type
            sizeof(RTAG) +      //anih tag
            PADUP(sizeof(TAniCurHeader)) +
            sizeof(RTAG);       //LIST tag (for fram list)
  IF (cbTitle > 0) OR (cbAuthor > 0) THEN
    begin
      IF cbTitle > 0 THEN
        begin
          Inc(cbTitle); // add in ASCIIZ terminator
          Inc(cbINFO, SizeOf(RTAG)); //INAM Tag
          Inc(cbINFO, PadUp(cbTitle));
        end;
      IF cbAuthor > 0 THEN
        begin
          Inc(cbAuthor); // add in ASCIIZ terminator
          Inc(cbINFO, SizeOf(RTAG)); //IART Tag
          Inc(cbINFO, PadUp(cbAuthor));
        end;
      inc(cbInfo, SizeOf(FourCC)); // INFO type
      Inc(cbFile, cbInfo);
      inc(cbFile, SizeOf(RTag)); // LIST tag
    end;
  IF fSeq THEN
    begin
      Inc(cbFile, SizeOf(RTag)); // seq tag
      Inc(cbFile, fSequences.Count*SizeOf(DWORD));
    end;
  IF fRate THEN
    begin
      Inc(cbFile, SizeOf(RTag)); // rate tag
      Inc(cbFile, fJifRates.Count*SizeOf(DWORD));
    end;
  // Now we have all the structures built in memory, it's time to
  // write them out in RIFF ACON format!
  Tag.ckID := FOURCC_RIFF;
  Tag.ckSize := cbFile;
  IF NOT WriteTag(Stream, Tag) THEN FErrOut;
  IF NOT WriteType(Stream, FOURCC_ACON) THEN FErrOut;
  // If title or author data exists, write a LIST type
  // block to hold them
  IF (cbTitle > 0) OR (cbAuthor > 0) THEN
    begin
      Tag.ckID := FOURCC_LIST;
      Tag.ckSize := cbInfo;
      IF NOT WriteTag(Stream, Tag) THEN FErrOut;
      IF NOT WriteType(Stream, FOURCC_INFO) THEN FErrOut;
      // If title exists, write an INAM block
      IF cbTitle > 0 THEN
        begin
          Tag.ckID := FOURCC_INAM;
          tag.ckSize := cbTitle;
          IF NOT WriteTagString(Stream, Tag, fTitle) THEN FErrOut;
        end;
      // IF author exists write an IART block
      IF cbAuthor > 0 THEN
        begin
          Tag.ckID := FOURCC_IART;
          tag.ckSize := cbAuthor;
          IF NOT WriteTagString(Stream, Tag, fAuthor) THEN FErrOut;
        end;
    end;
  Tag.ckID   := FOURCC_anih;
  Tag.ckSize := SizeOf(fheader);
  IF NOT WriteTagData(Stream, Tag, fHeader) THEN FErrOut;
  // If there's a list of separate rates for the steps, write
  // it in a Rate block.
  IF fRate THEN
    begin
      Tag.ckID := FourCC_Rate;
      Tag.ckSize := fJifRates.Count*SizeOf(DWORD);
      IF NOT WriteTag(Stream, Tag) THEN FErrOut;
      FOR i := 0 TO fJifRates.Count-1 DO
        IF NOT WriteType(Stream, DWORD(fJifRates[i])) THEN FErrOut;
    end;
  // If there's a sequence list to match each step to a frame,
  // write it in a _seq block
  IF fSeq THEN
    begin
      Tag.ckID := FourCC_Seq;
      Tag.ckSize := fSequences.Count*SizeOf(DWORD);
      IF NOT WriteTag(Stream, Tag) THEN FErrOut;
      FOR i := 0 TO fSequences.Count-1 DO
        IF NOT WriteType(Stream, DWORD(fSequences[i])) THEN FErrOut;
    end;
  // Write a LIST block to hold the frames
  Tag.ckID := FOURCC_LIST;
  Tag.ckSize := cbFram;
  IF NOT WriteTag(Stream, Tag) THEN FErrOut;
  IF NOT WriteType(Stream, FOURCC_Fram) THEN FErrOut;
  FOR i := 0 TO fFrams.Count-1 DO
    begin
      // Write an ICON block for each frame
      Tag.ckID := FOURCC_ICON;
      Tag.ckSize := TInternalCur(fFrams[i]).DataSize;
      WriteTag(Stream, Tag);
      TInternalCur(fFrams[i]).SaveToStream(Stream);
    end;
end;

procedure TAniCursor.LoadFromFile(const fileName: String);
VAR Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TAniCursor.SaveToFile(const fileName: String);
VAR Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmCreate);
  try
    SaveToStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TAniCursor.AddImage(T: TInternalCur);
VAR TII : TInternalCur;
begin
  fJifRates.Add(Pointer(T.Tag));
  TII     := TInternalCur.Create(T);
  TII.Tag := 0;
  fFrams.Add(TII);
  fSequences.Add(Pointer(fSequences.Count));
end;

procedure TAniCursor.PrepareToSave(vHotX, vHotY : Word);
VAR
  J, N, M, P : Integer;
  Any, Match : Boolean;
begin
  fHeader.cSteps  := fFrams.Count;
  fHeader.cFrames := fFrams.Count;
  Any             := False;
  J               := Integer(fJifRates[0]);
  FOR N := 1 TO fJifRates.Count-1 DO
    IF Integer(fJifRates[N]) <> J THEN
      begin
        Any := True;
        Break;
      end;
  IF NOT Any THEN // IF all the rates are the same, discard the list
    begin
      fJifRates.Clear;
      fHeader.cjifRate := J;
    end;
  Any := False;
  // The sequence list initially has the same number of items as
  // the Frames list, and item "n" indexes the "nth" frame. If
  // every frame is unique, discard the list. Otherwise, if frame
  // M matches earlier frame N, discard frame M and adjust the
  // sequence list
  FOR N := fHeader.cSteps-1 DOWNTO 1 DO
    begin
      Match :=False;
      FOR M := 0 TO N-1 DO
        IF TInternalCur(fFrams[N]).Same(TInternalCur(fFrams[M])) THEN
          begin
            Any           := True;
            Match         := True;
            fSequences[N] := Pointer(M);
            // When we remove a frame, we must update all sequence
            // indices that point to *later* frames
            FOR P := N+1 TO fSequences.Count-1 DO
              IF Integer(fSequences[P]) > N THEN
                fSequences[P] := Pointer(Integer(FSequences[P])-1);
            fFrams.Delete(N);
            Dec(fheader.cFrames);
            Break;
          end;
      IF Match THEN Continue;
    end;
  IF Any THEN
    fHeader.cflags := fHeader.cFlags OR AF_SEQUENCE
  ELSE
    fSequences.Clear;
  FOR N := 0 TO fFrams.Count-1 DO
    WITH TInternalCur(fFrams[N]) DO
      begin
        HotX := vHotX;
        HotY := vHotY;
      end;
end;

procedure TAniCursor.FErrIn;
begin
  Raise(Exception.Create('Error reading .ANI file'));
end;

procedure TAniCursor.FErrOut;
begin
  Raise(Exception.Create('Error writing .ANI file'));
end;

function TAniCursor.GetFrameCount: Integer;
begin
  Result := fFrams.Count;
end;

function TAniCursor.GetHotX: WORD;
begin
  IF fFrams.Count > 0 THEN
    Result := TInternalCur(fFrams[0]).HotX
  ELSE Result := 0;
end;

function TAniCursor.GetHotY: WORD;
begin
  IF fFrams.Count > 0 THEN
    Result := TInternalCur(fFrams[0]).HotY
  ELSE Result := 0;
end;

function TAniCursor.GetIsPal16: Boolean;
begin
  IF fFrams.Count > 0 THEN
    Result := TInternalCur(fFrams[0]).Is16color
  ELSE Result := True;
end;

function TAniCursor.GetRateCount: Integer;
begin
  Result := fJifRates.Count;
end;

function TAniCursor.GetSeqCount: Integer;
begin
  Result := fSequences.Count;
end;

function TAniCursor.GetPalette: TPalArray;
begin
  IF fFrams.Count > 0 THEN
    Result := TInternalCur(fFrams[0]).Palette
  ELSE FillChar(Result, SizeOf(Result), 0);
end;

function TAniCursor.GetFrame(Index: Integer): TInternalCur;
begin
  IF fSequences.Count = 0 THEN
    begin
      IF (Index < 0) or (Index >= fFrams.Count) THEN
        Raise(Exception.Create('List index out of bounds'));
      Result := TInternalCur(fFrams[Index]);
    end
  ELSE
    begin
      IF (Index < 0) or (Index >= fSequences.Count) THEN
        Raise(Exception.Create('List index out of bounds'));
      Result := TInternalCur(fFrams[Sequences[Index]]);
    end;
end;

function TAniCursor.GetRate(Index: Integer): DWORD;
begin
  IF fJifRates.Count = 0 THEN
    Result := fHeader.cjifRate
  ELSE
    begin
      IF (Index < 0) or (Index >= fJifRates.Count) THEN
        Raise(Exception.Create('List index out of bounds'));
      Result := DWORD(fJifRates[Index]);
    end;
end;

function TAniCursor.GetSequence(Index: Integer): DWORD;
begin
  IF fSequences.Count= 0 THEN
    Result := Index
  ELSE
    begin
      IF (Index < 0) or (Index >= fSequences.Count) THEN
        Raise(Exception.Create('List index out of bounds'));
      Result := DWORD(fSequences[Index]);
    end;
end;

end.
